{************************************************}
{                                                }
{   Turbo Pascal for Windows                     }
{   Icon Viewer Demo program                     }
{   Copyright (c) 1991 by Borland International  }
{                                                }
{   by Danny Thorpe                              }
{************************************************}

program IconView;

{$R Iconview.res}

uses WinTypes, WinProcs, WinDos, WObjects, Strings, iconfile;

const
  App_Name =  'IconView';
  MagnifierSize = 200;
  MagnifierRect : TRect = (Left  : 4;
                           Top   : 4;
                           Right : MagnifierSize-4;
                           Bottom: MagnifierSize-4);
  idIcon = 101;
  idDList = 102;
  idPathStatic = 103;

type

{ TIconViewApp is the application object that everything else 'lives in.'
  Note how the MessageLoop method has been changed to support processing
  when the rest of Windows is idle.  This allows for 'background' processing
  that runs as fast as Windows has time for - generally much faster than
  a Windows timer message can deliver.  A timer message gives your app
  control no sooner than the timer interval:  an idle method gives your
  app maximum cpu resources with minimal impact on system response time.
  Just be careful not to do too much in each pass through the idle method,
  or you will slow the whole system down.
}

  TIconViewApp = object(TApplication)
    procedure InitMainWindow; virtual;
    procedure MessageLoop; virtual;
    procedure Idle;
  end;

  PIconCollection = ^TIconCollection;
  TIconCollection = object(TCollection)
    procedure FreeItem(Item: Pointer);  virtual;
    end;

  PCanvas = ^TCanvas;
  PMainIconWindow = ^TMainIconWindow;

{ TMainIconWindow is our main window and is responsible for running the whole
  show.  It owns three child windows:  a TStatic that says 'Directory', a
  listbox of directories, and a Canvas window for showing the icons.  This
  main window paints the magnified bitmap of the selected icon in its top
  left corner.

  Some interesting areas to look at are:
  - How this TWindow manages to react to Enter keypresses (which Windows
    does not send as wm_KeyDown messages)
  - How child windows are resized proportional to the main window (see WMSize)
  - How the window's minimum size is controlled (see WMGetMinMaxInfo)
  - Processing in the Idle method, for 'background' processing
}
  TMainIconWindow = object(TWindow)
    PathName: PChar;
    FileList: PStrCollection;
    IdleFileLoadIndex: Integer;
    IdleMagnifierIndex: Integer;
    Canvas: PCanvas;
    DList: PListbox;
    MagColor, MagMono: HBitmap;
    BitInfo: TBitMap;
    constructor Init(ATitle: PChar);
    destructor Done; virtual;
    procedure ClearSettings;
    procedure DeleteBitmaps;
    procedure EnterKey(var Msg: TMessage); virtual cm_First + idOK;
    function  GetClassName: PChar; virtual;
    procedure GetWindowClass(var AWndClass: TWndClass); virtual;
    procedure HandleDList(var Msg: TMessage); virtual id_First + idDList;
    procedure Idle;
    procedure LoadFileNames;
    procedure Paint(PaintDC: HDC; var PaintInfo: TPaintStruct); virtual;
    procedure SetCaption;
    procedure SetupWindow; virtual;
    procedure WMGetMinMaxInfo(var Msg: TMessage); virtual wm_First + wm_GetMinMaxInfo;
    procedure WMSize(var Msg: TMessage);  virtual wm_First + wm_Size;
    procedure WMSysCommand(var Msg: TMessage); virtual wm_First + wm_SysCommand;
  end;



{ The Canvas window actually owns the icons that are created from the ico
  files and is responsible for displaying the icons, scrolling the display,
  and processing arrow keys and mouse clicks to select an icon for magnification.
}
  TCanvas = object(TWindow)
    IconList: PIconCollection;
    Selected: Integer;
    UpdateMagnifier : Boolean;
    constructor Init(AParent: PWindowsObject; ATitle: Pchar);
    destructor Done; virtual;
    procedure AddIcon(Icon: HIcon);
    procedure CalcClipRect(Index: Integer; var R: TRect);
    procedure ClearIcons;
    procedure DrawSelection;
    procedure EraseSelection;
    function  HasIcons: Boolean;
    procedure Paint(PaintDC: HDC; var PaintInfo: TPaintStruct); virtual;
    procedure WMGetDlgCode(var Msg: TMessage); virtual wm_First + wm_GetDlgCode;
    procedure WMKeyDown(var Msg: TMessage); virtual wm_First + wm_KeyDown;
    procedure WMKeyUp(var Msg: TMessage); virtual wm_First + wm_KeyUp;
    procedure WMKillFocus(var Msg: TMessage); virtual wm_First + wm_KillFocus;
    procedure WMLButtonDown(var Msg: TMessage); virtual wm_First + wm_LButtonDown;
    procedure WMSetFocus(var Msg: TMessage); virtual wm_First + wm_SetFocus;
    procedure WMSize(var Msg: TMessage);  virtual wm_First + wm_Size;
    end;


function Min(L1, L2: Longint): Longint;
begin
  if L1 < L2 then
    Min := L1
  else
    Min := L2;
end;


{ -----------  TIconViewApp  ------------- }

procedure TIconViewApp.InitMainWindow;
begin
  MainWindow := New(PMainIconWindow, Init(App_name));
end;

{ Override the default application message loop so we can do things when
  the system is idle (when there are no messages to be processed)  }

procedure TIconViewApp.MessageLoop;
var
  Message: TMsg;
  Quit: Boolean;
begin
  Quit := False;
  while not Quit do
  begin
    if PeekMessage(Message, 0, 0, 0, pm_Remove) then
    begin
      Quit := (Message.Message = wm_Quit);
      if (not Quit) and (not ProcessAppMsg(Message)) then
      begin
        TranslateMessage(Message);
        DispatchMessage(Message);
      end
    end
    else
      Idle;
  end;
  Status := Message.WParam;
end;

procedure TIconViewApp.Idle;
begin
  PMainIconWindow(MainWindow)^.Idle;
end;


{ ----------- THandleCollection ------------}

procedure TIconCollection.FreeItem(Item: Pointer);
begin
  if THandle(Item) <> 0 then
    DestroyIcon(THandle(Item));  { DONT use DeleteObject !}
end;


{ ----------- TMainIconWindow --------------}

constructor TMainIconWindow.Init(ATitle: PChar);
var
  P : PStatic;
begin
  TWindow.Init(nil, ATitle);
  Attr.Menu := LoadMenu(HInstance, App_Name);
  FileList := New(PStrCollection, Init(16, 16));
  IdleFileLoadIndex := 0;
  MagMono := 0;
  MagColor := 0;
  IdleMagnifierIndex := -1;

  Canvas := new(PCanvas, Init(@Self, 'Icon Canvas'));
  P := new(PStatic, Init(@Self, idPathStatic, 'Directories', 4, MagnifierSize + 4,
                         MagnifierSize - 4, 15, 13));

  GetMem(PathName, 100);
  GetCurDir(PathName, 0);   { current directory of current drive }
  if PathName[strlen(PathName)] <> '\' then
    StrCat(PathName, '\*.ico')
  else
    StrCat(PathName, '*.ico');

  { Listbox width and height will be fixed up in WMSize }
  DList := new(PListBox, Init(@Self, idDList, 4, MagnifierSize + 20, 100, 100));
  EnableKBHandler;

end;


destructor TMainIconWindow.Done;
begin    { Since Pathstatic, Canvas, and DList are child windows, they will }
         { be disposed of automatically.  }
  TWindow.Done;
  FreeMem(PathName, 100);
  Dispose(FileList, Done);
  DeleteBitmaps;
end;


procedure TMainIconWindow.ClearSettings;
begin
  Canvas^.ClearIcons;
  DeleteBitmaps;
  IdleMagnifierIndex := -1;
  IdleFileLoadIndex := 0;   { Tell the idle proc to get to work loading those icons }
  InvalidateRect(HWindow, @MagnifierRect, True);
end;


procedure TMainIconWindow.DeleteBitmaps;
begin
  if MagColor <> 0 then
  begin
    if not DeleteObject(MagColor) then
      MessageBeep(0);
    MagColor := 0;
  end;

  if MagMono <> 0 then
  begin
    if not DeleteObject(MagMono) then
      MessageBeep(0);
    MagMono := 0;
  end;
end;


{ When a dialog or a window with EnableKBHandler set recieves an Enter key,
  Windows converts that into a message that looks like an Ok button click
  (even when we don't have an ok button).  The window never receives a
  wm_KeyDown message containing vk_Return key code.  }
procedure TMainIconWindow.EnterKey(var Msg: TMessage);
begin
  { if the Directory listbox is focused ...}
  if GetFocus = DList^.HWindow then
  begin
    if DList^.GetSelIndex < 0 then   { If nothing is selected in the listbox }
      DList^.SetSelIndex(0);         { then select the first item }
    Msg.LparamLo := DList^.HWindow;
    Msg.LParamHi := lbn_DblClk;      { Repackage the message as a double click }
    HandleDlist(Msg);                { and give it to our list box message handler. }
  end;
end;


function TMainIconWindow.GetClassName: PChar;
begin
  GetClassName := 'Turbo Icon Viewer';
end;


{ Associate an icon with the main window }
procedure TMainIconWindow.GetWindowClass(var AWndClass: TWndClass);
begin
  TWindow.GetWindowClass(AWndClass);
  AWndClass.HIcon := LoadIcon(hInstance, PChar(idIcon));
end;


{ Process messages from the directory list box }
procedure TMainIconWindow.HandleDList(var Msg: TMessage);
var Temp : array [0..fsPathName] of char;
begin
  case Msg.LParamHi of
  lbn_DblClk:
    begin
      { User has double clicked on a directory. }
      SetCursor(LoadCursor(hInstance, idc_Wait));
      DlgDirSelect(HWindow, PathName, idDList);  { Get the directory name }
      StrCat(PathName, '*.ico');                 { Add *.ico to it }
      FileExpand(PathName, PathName);            { Expand it to a full path name }
      SetCaption;                                { Update the main window title }
      LoadFileNames;                             { Load the file names }
      ClearSettings;

      { Now tell Windows to fill the list box with drive letters and
        subdirectories of the selected directory.
        DlgDirList modifies the second parameter.  We don't want PathName modified.
      }
      DlgDirList(HWindow, StrCopy(Temp, PathName), idDList, 0, $C010);
      SetCursor(LoadCursor(hInstance, idc_Arrow));
    end;
  lbn_KillFocus:
    SendMessage(Msg.LParamLo, lb_SetCurSel, Word(-1), 0);
  lbn_SetFocus:
    SendMessage(Msg.LParamLo, lb_SetCurSel, 0, 0);
  end;
end;


{ Perform 'background'  processing here:
  - Loading of bitmaps for Magnifier display
  - Loading ico files from current directory }
procedure TMainIconWindow.Idle;
begin
  if (HWindow = 0) or (Canvas = nil) then Exit;
  if Canvas^.HasIcons and
     Canvas^.UpdateMagnifier and
     (IdleMagnifierIndex <> Canvas^.Selected) then
  begin
    DeleteBitmaps;
    LoadIconFileBitmaps(FileList^.At(Canvas^.Selected), MagColor, MagMono, BitInfo);
    IdleMagnifierIndex := Canvas^.Selected;
    InvalidateRect(HWindow, @MagnifierRect, False);
  end
  else    { give icon loading lower priority than magnifier loading }
    if (IdleFileLoadIndex < FileList^.Count) then
    begin
      Canvas^.AddIcon(LoadIconFile(FileList^.At(IdleFileLoadIndex)));
      Inc(IdleFileLoadIndex);
    end;
end;


{ Get all the *.ico file names from the directory specified in Pathname
  and put them into a string collection.  }
procedure TMainIconWindow.LoadFileNames;
var Temp: array [0..fsPathName] of char;
    Name: array [0..fsFileName] of char;
    Ext: array [0..fsExtension] of char;
    DirInfo: TSearchRec;
begin
  FileList^.FreeAll;
  FindFirst(PathName, faAnyFile, DirInfo);
  while DosError = 0 do
  begin
    FileSplit(PathName, Temp, Name, Ext);
    StrCat(Temp, DirInfo.Name);
    FileList^.Insert(StrNew(Temp));
    FindNext(DirInfo);
  end;
end;


{ Paint the main window - all we need to do here is paint the magnified
  icon bitmap, if we have one.  The child windows (Canvas and DList)
  will paint themselves as needed. }
procedure TMainIconWindow.Paint(PaintDC: HDC; var PaintInfo: TPaintStruct);
var OldBitmap1, OldBitmap2: HBitmap;
    Dest, Source: HDC;
begin
  if (MagColor <> 0) and (MagMono <> 0) then
    with MagnifierRect do
    begin
      { Build the stretched image of the icon in memory device contexts. }
      { We're using memdcs for speed and no screen flicker.  }
      Dest := CreateCompatibleDC(PaintDC);
      Source:= CreateCompatibleDC(PaintDC);
      OldBitmap1 := SelectObject(Dest, CreateCompatibleBitmap(PaintDC, Right, Bottom));
      BitBlt(Dest, 0, 0, Right, Bottom, 0, 0, 0, Whiteness);
      OldBitmap2 := SelectObject(Source, MagMono);
      StretchBlt(Dest, 0, 0, Right, Bottom,
                 Source, 0, 0, BitInfo.bmWidth, BitInfo.bmHeight, SrcAND);
      SelectObject(Source, MagColor);
      StretchBlt(Dest, 0, 0, Right, Bottom,
                 Source, 0, 0, BitInfo.bmWidth, BitInfo.bmHeight, SrcInvert);

      { Transfer our work to the screen }
      BitBlt(PaintDC, Left, Top, Right-Left, Bottom-Top, Dest, 0, 0, SrcCopy);

      { Put the original Dest bitmap back in, and delete the bitmap we created. }
      DeleteObject(SelectObject(Dest, OldBitmap1));
      SelectObject(Source, OldBitmap2);
      DeleteDC(Source);
      DeleteDC(Dest);
    end;
end;


{ Change the main window's title bar to show the selected path name. }
procedure TMainIconWindow.SetCaption;
var Caption: array[0..fsPathName] of char;
begin
  if HWindow <> 0 then
  begin
    StrCopy(Caption, App_Name);
    StrCat(Caption, ' - [');
    StrCat(Caption, PathName);
    StrCat(Caption, ']');
    SetWindowText(HWindow, Caption);
  end;
end;


procedure TMainIconWindow.SetupWindow;
var Temp: array [0..fsPathName] of char;
    Result : Word;
begin
  TWindow.SetupWindow;

  SetCaption;
  LoadFileNames;

  { DlgDirList modifies the second parameter.  We don't want PathName modified.}
  DlgDirList(HWindow, StrCopy(Temp, PathName), idDList, 0, $C010);

  if Canvas^.HasIcons then
    SetFocus(Canvas^.HWindow)
  else
    SetFocus(DList^.HWindow);
end;


{ Control how small the main window can be resized to.  Min tracking size
  is the 4th item (index 3) in the array of points that Windows sends us
  while the window is being resized.  See your Windows API ref.  }

procedure TMainIconWindow.WMGetMinMaxInfo(var Msg: TMessage);
type TMinMaxInfo = array [0..4] of TPoint;
     PMinMaxInfo = ^TMinMaxInfo;
var MinMaxInfo: PMinMaxInfo;
begin
  with PMinMaxInfo(Msg.LParam)^[3] do
  begin
    if X < MagnifierSize + 74 + 40 then
      X := MagnifierSize + 74 + 40;
    if Y < MagnifierSize + 110 then
      Y := MagnifierSize + 110;
  end;
end;


{ When the main window is resized, tell the child windows to resize also,
  and tell them what size to change to. }

procedure TMainIconWindow.WMSize(var Msg: TMessage);
begin
  TWindow.WMSize(Msg);
  if (Msg.WParam = SizeNormal) or (Msg.WParam = SizeFullScreen) then
  begin
    if (Canvas <> nil) and (Canvas^.HWindow <> 0) then
      MoveWindow(Canvas^.HWindow, MagnifierSize+1, -1,
                                  Msg.LParamLo - MagnifierSize + 1,
                                  Msg.LParamHi + 2, True);
    if (DList <> nil) and (DList^.HWindow <> 0) then
      MoveWindow(DList^.HWindow, 4, MagnifierSize + 20,
                                    MagnifierSize - 8,
                                    Msg.LParamHi - (MagnifierSize + 30), True);
  end;
end;

{ Save the focus state before the main window is minimized, otherwise
  Windows will lose track of who had focus last. }
procedure TMainIconWindow.WMSysCommand(var Msg: TMessage);
begin
  if Msg.WParam = sc_Minimize then
    FocusChildHandle := GetFocus;
  DefWndProc(Msg);
end;




{------------------ TCanvas ---------------}
constructor TCanvas.Init(AParent: PWindowsObject; ATitle: PChar);
begin
  TWindow.Init(AParent, ATitle);
  IconList := New(PIconCollection, Init(16,16));

  { The tabstop style lets us tab from the list box into the canvas window. }
  Attr.Style := ws_Child or ws_Border or ws_Visible or ws_VScroll or ws_TabStop;

  { Initiallize scroller }
  Scroller := new(PScroller, Init(@Self, 0, 74, 0, 74));
  Selected := 0;
  UpdateMagnifier := True;
end;

destructor TCanvas.Done;
begin
  TWindow.Done;
  Dispose(IconList, Done);
end;


procedure TCanvas.AddIcon(Icon: HIcon);
var I, Width: integer;
    R: TRect;
    TotalLines, ShowableLines: Integer;
begin
  I := IconList^.Count;

  if I = 0 then  { Clear the 'no files' message from the screen }
    InvalidateRect(HWindow, nil, True);

  { Typecast is necessary to force the collection to accept a handle }
  IconList^.AtInsert(I, Pointer(Icon));

  { Calculate rectangle where this icon will be drawn, to minimize paint time }
  Width := (Attr.W - 10) div 74;
  TotalLines := I div Width;
  ShowableLines := (Attr.H -10) div 74;

  CalcClipRect(I, R);
  InvalidateRect(HWindow, @R, False);

  if TotalLines > ShowableLines then
  begin
    Scroller^.SetRange(0, TotalLines - ShowableLines + 1);
    SetScrollPos(HWindow, sb_Vert, Scroller^.YPos, True);
  end;
end;


{ Given an index, calculate the rectangle around the icon at that index }
procedure TCanvas.CalcClipRect(Index: Integer; var R: TRect);
var Width: Integer;
begin
  Width := (Attr.W - 10) div 74;
  R.Left := 74 * (Index mod Width) + 10;
  R.Right := R.Left + 74;
  R.Top :=  74 * ((Index div Width) - Scroller^.YPos) + 10;
  R.Bottom := R.Top + 74;
end;


{ Destroy all the icons, reset the scrollbar, and clear the window. }
procedure TCanvas.ClearIcons;
begin
  IconList^.FreeAll;
  Scroller^.SetRange(0,0);
  Selected := 0;
  InvalidateRect(HWindow, nil, TRUE);
end;


{ Invalidate the area around the selected icon so that the Paint method will
  draw it. }
procedure TCanvas.DrawSelection;
var DC: HDC;
    R: TRect;
    OldBrush: HBrush;
begin
  CalcClipRect(Selected, R);
  Dec(R.Left,2);
  Dec(R.Top,2);
  InvalidateRect(HWindow, @R, True);
end;


{ Same as DrawSelection, but it reads better as a separate function. Since the
selection is currently an inversion operation, drawing the selection once will
highlight the icon (white becomes black), and drawing the selection again will
restore the icon (black becomes white) }
procedure TCanvas.EraseSelection;
begin
  DrawSelection;
end;


function TCanvas.HasIcons: Boolean;
begin
  HasIcons := (IconList <> nil) and (IconList^.Count > 0);
end;


procedure TCanvas.Paint(PaintDC: HDC; var PaintInfo: TPaintStruct);
var x, y, I: integer;
    Dest, Icon, ScrollClip: TRect;
    OldBrush: HBrush;
    OldFont: HFont;
    OldTextColor: Longint;
    OldBkMode: Integer;
    Metrics: TTextMetric;
    FullName: PChar;
    Dir: array [0..fsDirectory] of Char;
    Name: array [0..fsFileName] of Char;
    Ext: array [0..fsExtension] of Char;

  procedure DrawTheIcon(Item: Pointer);  far;
  begin
    { Calculate the rectangle around the current icon }
    with Icon do
    begin
      Left := x;
      Top := y;
      Right := x + 74;
      Bottom := y + 74;
    end;

    { If the current icon is not part of the area to be repainted, then
      don't bother.  We save time by not drawing something that would be
      clipped anyway.  }
    if WordBool(IntersectRect(Dest, Icon, ScrollClip)) then
    begin
      FullName := PMainIconWindow(Parent)^.FileList^.At(I);
      FileSplit(FullName, Dir, Name, Ext);    { Extract name }

      { If the current icon is the selected icon, and the Canvas window is
        focused, then we draw the selected icon with a black background
        and white text.  }
      if (I = Selected) and (GetFocus = HWindow) then
      begin
        OldBrush := SelectObject(PaintDC, GetStockObject(Black_Brush));

        Rectangle(PaintDC, Icon.Left-2, Icon.Top-2, Icon.Right, Icon.Bottom);

        OldBkMode := SetBKMode(PaintDC, Transparent);
        OldTextColor := SetTextColor(PaintDC, RGB($FF, $FF, $FF));  {white}
        TextOut(PaintDC, x, y, Name, StrLen(Name));
        SetTextColor(PaintDC, OldTextColor);
        SetBkMode(PaintDC, OldBkMode);

        DrawIcon(PaintDC, x, y + Metrics.tmHeight, THandle(Item));

        SelectObject(PaintDC, OldBrush);
      end
      else
      begin
        TextOut(PaintDC, x, y, Name, StrLen(Name));
        DrawIcon(PaintDC, x, y + Metrics.tmHeight, THandle(Item));
      end;
    end;

    { Increment our counters for the next icon }
    Inc(X, 74);
    Inc(I);
    if X > (Attr.W - 74) then
    begin
      Inc(Y, 74);
      X := 10;
    end;
  end;

begin
  X := 10;
  Y := 10;
  I := 0;

  { Adjust the clipping rectangle to the scroller origin. }
  with PaintInfo.rcPaint do
  begin
    ScrollClip.Top := Top + (Scroller^.YPos * 74 + 10);
    ScrollClip.Bottom := Bottom + (Scroller^.YPos * 74 + 10);
    ScrollClip.Left := Left;
    ScrollClip.Right := Right;
  end;

  if IconList^.Count > 0 then
  begin
    OldFont := SelectObject(PaintDC, GetStockObject(System_Font));
    GetTextMetrics(PaintDC, Metrics);   { Metrics contains the font height }
    IconList^.ForEach(@DrawTheIcon);
    SelectObject(PaintDC, OldFont);
  end
  else
    TextOut(PaintDC, x, y, 'No *.ico files in this directory.', 33);
end;


procedure TCanvas.WMGetDlgCode(var Msg: TMessage);
begin
  Msg.Result := dlgc_WantArrows;   { tell'em we want to use the arrow keys }
end;


{ Move the icon selection around using the arrow keys }
procedure TCanvas.WMKeyDown(var Msg: TMessage);
var Width, ShowableRows, SelectedRow, NewSelection: Integer;
begin
  Width := (Attr.W - 10) div 74;
  NewSelection := Selected;
  case Msg.WParam of
    vk_Left : if Selected > 0 then
                NewSelection := Selected - 1;
    vk_Right: if Selected < (IconList^.Count -1) then
                NewSelection := Selected + 1;
    vk_Up   : if Selected - Width >= 0 then
                NewSelection := Selected - Width;
    vk_Down : if Selected + Width < IconList^.Count then
                NewSelection := Selected + Width;
  end;

  if NewSelection <> Selected then
  begin
    EraseSelection;
    Selected := NewSelection;
    ShowableRows := (Attr.H - 10) div 74;
    SelectedRow := Selected div Width;
    if SelectedRow < Scroller^.YPos then
      Scroller^.ScrollBy(0, SelectedRow - Scroller^.YPos)  {negative scroll}
    else
      if (Scroller^.YPos + ShowableRows -1) < SelectedRow then {positive scroll}
        Scroller^.ScrollBy(0, SelectedRow - (Scroller^.YPos + ShowableRows)+1);
    DrawSelection;              { Don't update the icon magnifier }
    UpdateMagnifier := False;   { until we get a KeyUp (for speed reasons)}
  end;

  DefWndProc(Msg);
end;


procedure TCanvas.WMKeyUp(var Msg: TMessage);
begin
  if Msg.WParam in [vk_Up, vk_Down, vk_Left, vk_Right] then
    UpdateMagnifier := True;     { ok, now let the magnifier be updated }
  DefWndProc(Msg);
end;


procedure TCanvas.WMKillFocus(var Msg: TMessage);
var R: TRect;
begin
  EraseSelection;
  DefWndProc(Msg);
end;


{ Select the icon the user clicked on }
procedure TCanvas.WMLButtonDown(var Msg: TMessage);
var P: TPoint;
    Width: Integer;
    I: Integer;
    R: TRect;
begin
  SetFocus(HWindow);

  P := TPoint(Msg.LParam);

  { figure out which icon they clicked on, if any}
  if (P.X > 10) and (P.Y > 10) then
  begin
    Width := ((Attr.W - 10) div 74);           { number of icons per row }
    I := ((P.X - 10) div 74) +
         ((((P.Y - 10) div 74) + Scroller^.YPos) * Width);
    if (I >= 0) and (I < IconList^.Count) then
    begin
      EraseSelection;
      Selected := I;
      CalcClipRect(Selected, R);
      if R.Top < 0 then                { If they clicked on a partially }
        Scroller^.ScrollBy(0, -1)      { displayed icon, scroll it into view.}
      else
        if R.Bottom > Attr.H then
          Scroller^.ScrollBy(0, 1);
      DrawSelection;
    end;
  end;

  TWindow.WMLButtonDown(Msg);
end;


procedure TCanvas.WMSetFocus(var Msg: TMessage);
var R: TRect;
begin
  DrawSelection;
  DefWndProc(Msg);
end;


{ Rescale the scrollbar range to reflect the possibly changed number of
  icons per row/ number of rows of icons.  }
procedure TCanvas.WMSize(var Msg: TMessage);
var Width: Integer;
    TotalRows: Integer;
    ShowableRows: Integer;
begin
  TWindow.WMSize(Msg);
  Width := ((Msg.LParamLo - 10) div 74);
  TotalRows := (IconList^.Count -1) div Width;
  ShowableRows := (Msg.LParamHi - 10) div 74;
  Scroller^.SetRange(0, TotalRows - ShowableRows + 1);
end;


{ Declare a variable of type TIconViewApp }

var
  IconApp: TIconViewApp;

{ Run the IconViewApp }

begin
  IconApp.Init(App_Name);
  IconApp.Run;
  IconApp.Done;
end.
